@@ -1,35 +1,37 @@
 # frozen_string_literal: true
 
 # Released under the MIT License.
-# Copyright, 2019-2023, by Samuel Williams.
+# Copyright, 2019-2024, by Samuel Williams.
 # Copyright, 2019, by Brian Morearty.
 # Copyright, 2020, by Bruno Sutic.
+# Copyright, 2023-2024, by Thomas Morgan.
+# Copyright, 2024, by Anton Zhuravsky.
 
-require 'protocol/http/headers'
+require "protocol/http/headers"
 
-require_relative 'reason'
-require_relative 'error'
+require_relative "reason"
+require_relative "error"
 
-require_relative 'body/chunked'
-require_relative 'body/fixed'
-require_relative 'body/remainder'
-require 'protocol/http/body/head'
+require_relative "body/chunked"
+require_relative "body/fixed"
+require_relative "body/remainder"
+require "protocol/http/body/head"
 
-require 'protocol/http/methods'
+require "protocol/http/methods"
 
 module Protocol
 	module HTTP1
-		CONTENT_LENGTH = 'content-length'
+		CONTENT_LENGTH = "content-length"
 		
-		TRANSFER_ENCODING = 'transfer-encoding'
-		CHUNKED = 'chunked'
+		TRANSFER_ENCODING = "transfer-encoding"
+		CHUNKED = "chunked"
 		
-		CONNECTION = 'connection'
-		CLOSE = 'close'
-		KEEP_ALIVE = 'keep-alive'
+		CONNECTION = "connection"
+		CLOSE = "close"
+		KEEP_ALIVE = "keep-alive"
 		
-		HOST = 'host'
-		UPGRADE = 'upgrade'
+		HOST = "host"
+		UPGRADE = "upgrade"
 		
 		# HTTP/1.x request line parser:
 		TOKEN = /[!#$%&'*+\-\.\^_`|~0-9a-zA-Z]+/.freeze
@@ -43,23 +45,94 @@ module Protocol
 		VALID_FIELD_NAME = /\A#{FIELD_NAME}\z/.freeze
 		VALID_FIELD_VALUE = /\A#{FIELD_VALUE}\z/.freeze
 		
+		DEFAULT_MAXIMUM_LINE_LENGTH = 8192
+		
 		class Connection
 			CRLF = "\r\n"
 			HTTP10 = "HTTP/1.0"
 			HTTP11 = "HTTP/1.1"
 			
-			def initialize(stream, persistent = true)
+			def initialize(stream, persistent: true, state: :idle, maximum_line_length: DEFAULT_MAXIMUM_LINE_LENGTH)
 				@stream = stream
 				
 				@persistent = persistent
+				@state = state
 				
 				@count = 0
+				
+				@maximum_line_length = maximum_line_length
 			end
 			
 			attr :stream
 			
 			# Whether the connection is persistent.
-			attr :persistent
+			# This determines what connection headers are sent in the response and whether
+			# the connection can be reused after the response is sent.
+			# This setting is automatically managed according to the nature of the request
+			# and response.
+			# Changing to false is safe.
+			# Changing to true from outside this class should generally be avoided and,
+			# depending on the response semantics, may be reset to false anyway.
+			attr_accessor :persistent
+			
+			# The current state of the connection.
+			#
+			# ```
+			#                          ┌────────┐
+			#                          │        │
+			# ┌───────────────────────►│  idle  │
+			# │                        │        │
+			# │                        └───┬────┘
+			# │                            │
+			# │                            │ send request /
+			# │                            │ receive request
+			# │                            │
+			# │                            ▼
+			# │                        ┌────────┐
+			# │                recv ES │        │ send ES
+			# │           ┌────────────┤  open  ├────────────┐
+			# │           │            │        │            │
+			# │           ▼            └───┬────┘            ▼
+			# │      ┌──────────┐          │           ┌──────────┐
+			# │      │   half   │          │           │   half   │
+			# │      │  closed  │          │ send R /  │  closed  │
+			# │      │ (remote) │          │ recv R    │ (local)  │
+			# │      └────┬─────┘          │           └─────┬────┘
+			# │           │                │                 │
+			# │           │ send ES /      │       recv ES / │
+			# │           │ close          ▼           close │
+			# │           │            ┌────────┐            │
+			# │           └───────────►│        │◄───────────┘
+			# │                        │ closed │
+			# └────────────────────────┤        │
+			#         persistent       └────────┘
+			# ```
+			#
+			# - `ES`: the body was fully received or sent (end of stream).
+			# - `R`: the connection was closed unexpectedly (reset).
+			#
+			# State transition methods use a trailing "!".
+			attr_accessor :state
+			
+			def idle?
+				@state == :idle
+			end
+			
+			def open?
+				@state == :open
+			end
+			
+			def half_closed_local?
+				@state == :half_closed_local
+			end
+			
+			def half_closed_remote?
+				@state == :half_closed_remote
+			end
+			
+			def closed?
+				@state == :closed
+			end
 			
 			# The number of requests processed.
 			attr :count
@@ -97,24 +170,63 @@ module Protocol
 				@stream.write("connection: upgrade\r\nupgrade: #{upgrade}\r\n")
 			end
 			
+			# Indicates whether the connection has been hijacked meaning its
+			# IO has been handed over and is not usable anymore.
+			# @return [Boolean] hijack status
+			def hijacked?
+				@stream.nil?
+			end
+			
 			# Effectively close the connection and return the underlying IO.
 			# @return [IO] the underlying non-blocking IO.
 			def hijack!
 				@persistent = false
-				stream = @stream
 				
-				@stream.flush
-				@stream = nil
-				
-				return stream
+				if stream = @stream
+					@stream = nil
+					stream.flush
+					
+					@state = :hijacked
+					self.closed
+					
+					return stream
+				end
+			end
+			
+			def close_read
+				@persistent = false
+				@stream&.close_read
+				self.receive_end_stream!
 			end
 			
 			# Close the connection and underlying stream.
-			def close
-				@stream&.close
+			def close(error = nil)
+				@persistent = false
+				
+				if stream = @stream
+					@stream = nil
+					stream.close
+				end
+				
+				unless closed?
+					@state = :closed
+					self.closed(error)
+				end
+			end
+			
+			def open!
+				unless @state == :idle
+					raise ProtocolError, "Cannot open connection in state: #{@state}!"
+				end
+				
+				@state = :open
+				
+				return self
 			end
 			
 			def write_request(authority, method, path, version, headers)
+				open!
+				
 				@stream.write("#{method} #{path} #{version}\r\n")
 				@stream.write("host: #{authority}\r\n")
 				
@@ -122,12 +234,29 @@ module Protocol
 			end
 			
 			def write_response(version, status, headers, reason = Reason::DESCRIPTIONS[status])
+				unless @state == :open or @state == :half_closed_remote
+					raise ProtocolError, "Cannot write response in state: #{@state}!"
+				end
+				
 				# Safari WebSockets break if no reason is given:
 				@stream.write("#{version} #{status} #{reason}\r\n")
 				
 				write_headers(headers)
 			end
 			
+			def write_interim_response(version, status, headers, reason = Reason::DESCRIPTIONS[status])
+				unless @state == :open or @state == :half_closed_remote
+					raise ProtocolError, "Cannot write interim response in state: #{@state}!"
+				end
+				
+				@stream.write("#{version} #{status} #{reason}\r\n")
+				
+				write_headers(headers)
+				
+				@stream.write("\r\n")
+				@stream.flush
+			end
+			
 			def write_headers(headers)
 				headers.each do |name, value|
 					# Convert it to a string:
@@ -148,15 +277,30 @@ module Protocol
 				end
 			end
 			
+			def readpartial(length)
+				@stream.readpartial(length)
+			end
+			
+			def read(length)
+				@stream.read(length)
+			end
+			
 			def read_line?
-				@stream.gets(CRLF, chomp: true)
+				if line = @stream.gets(CRLF, @maximum_line_length)
+					unless line.chomp!(CRLF)
+						# This basically means that the request line, response line, header, or chunked length line is too long.
+						raise LineLengthError, "Line too long!"
+					end
+				end
+				
+				return line
 			end
 			
 			def read_line
 				read_line? or raise EOFError
 			end
 			
-			def read_request
+			def read_request_line
 				return unless line = read_line?
 				
 				if match = line.match(REQUEST_LINE)
@@ -165,29 +309,62 @@ module Protocol
 					raise InvalidRequest, line.inspect
 				end
 				
+				return method, path, version
+			end
+			
+			def read_request
+				open!
+				
+				method, path, version = read_request_line
+				return unless method
+				
 				headers = read_headers
 				
 				@persistent = persistent?(version, method, headers)
 				
 				body = read_request_body(method, headers)
 				
+				unless body
+					self.receive_end_stream!
+				end
+				
 				@count += 1
 				
 				return headers.delete(HOST), method, path, version, headers, body
 			end
 			
-			def read_response(method)
+			def read_response_line
 				version, status, reason = read_line.split(/\s+/, 3)
 				
 				status = Integer(status)
 				
+				return version, status, reason
+			end
+			
+			private def interim_status?(status)
+				status != 101 and status >= 100 and status < 200
+			end
+			
+			def read_response(method)
+				unless @state == :open or @state == :half_closed_local
+					raise ProtocolError, "Cannot read response in state: #{@state}!"
+				end
+				
+				version, status, reason = read_response_line
+				
 				headers = read_headers
 				
 				@persistent = persistent?(version, method, headers)
 				
-				body = read_response_body(method, status, headers)
-				
-				@count += 1
+				unless interim_status?(status)
+					body = read_response_body(method, status, headers)
+					
+					unless body
+						self.receive_end_stream!
+					end
+					
+					@count += 1
+				end
 				
 				return version, status, reason, headers, body
 			end
@@ -202,13 +379,23 @@ module Protocol
 					if match = line.match(HEADER)
 						fields << [match[1], match[2]]
 					else
-						raise BadHeader, "Could not parse header: #{line.dump}"
+						raise BadHeader, "Could not parse header: #{line.inspect}"
 					end
 				end
 				
 				return HTTP::Headers.new(fields)
 			end
 			
+			def send_end_stream!
+				if @state == :open
+					@state = :half_closed_local
+				elsif @state == :half_closed_remote
+					self.close!
+				else
+					raise ProtocolError, "Cannot send end stream in state: #{@state}!"
+				end
+			end
+			
 			# @param protocol [String] the protocol to upgrade to.
 			def write_upgrade_body(protocol, body = nil)
 				# Once we upgrade the connection, it can no longer handle other requests:
@@ -229,6 +416,8 @@ module Protocol
 				end
 				
 				return @stream
+			ensure
+				self.send_end_stream!
 			end
 			
 			def write_tunnel_body(version, body = nil)
@@ -249,6 +438,8 @@ module Protocol
 				end
 				
 				return @stream
+			ensure
+				self.send_end_stream!
 			end
 			
 			def write_empty_body(body)
@@ -256,6 +447,8 @@ module Protocol
 				@stream.flush
 				
 				body&.close
+			ensure
+				self.send_end_stream!
 			end
 			
 			def write_fixed_length_body(body, length, head)
@@ -288,6 +481,8 @@ module Protocol
 				if chunk_length != length
 					raise ContentLengthError, "Wrote #{chunk_length} bytes, but content length was #{length} bytes!"
 				end
+			ensure
+				self.send_end_stream!
 			end
 			
 			def write_chunked_body(body, head, trailer = nil)
@@ -322,6 +517,8 @@ module Protocol
 				end
 				
 				@stream.flush
+			ensure
+				self.send_end_stream!
 			end
 			
 			def write_body_and_close(body, head)
@@ -341,7 +538,30 @@ module Protocol
 					end
 				end
 				
+				@stream.flush
 				@stream.close_write
+			ensure
+				self.send_end_stream!
+			end
+			
+			# The connection (stream) was closed. It may now be in the idle state.
+			def closed(error = nil)
+			end
+			
+			# Transition to the closed state.
+			#
+			# If no error occurred, and the connection is persistent, this will immediately transition to the idle state.
+			#
+			# @parameter error [Exxception] the error that caused the connection to close.
+			def close!(error = nil)
+				if @persistent and !error
+					# If there was no error, and the connection is persistent, we can reuse it:
+					@state = :idle
+				else
+					@state = :closed
+				end
+				
+				self.closed(error)
 			end
 			
 			def write_body(version, body, head = false, trailer = nil)
@@ -375,19 +595,33 @@ module Protocol
 				end
 			end
 			
+			def receive_end_stream!
+				if @state == :open
+					@state = :half_closed_remote
+				elsif @state == :half_closed_local
+					self.close!
+				else
+					raise ProtocolError, "Cannot receive end stream in state: #{@state}!"
+				end
+			end
+			
 			def read_chunked_body(headers)
-				Body::Chunked.new(@stream, headers)
+				Body::Chunked.new(self, headers)
 			end
 			
 			def read_fixed_body(length)
-				Body::Fixed.new(@stream, length)
+				Body::Fixed.new(self, length)
 			end
 			
 			def read_remainder_body
-				Body::Remainder.new(@stream)
+				@persistent = false
+				Body::Remainder.new(self)
 			end
 			
 			def read_head_body(length)
+				# We are not receiving any body:
+				self.receive_end_stream!
+				
 				Protocol::HTTP::Body::Head.new(length)
 			end
 			
@@ -395,15 +629,23 @@ module Protocol
 				read_remainder_body
 			end
 			
+			def read_upgrade_body
+				# When you have an incoming upgrade request body, we must be extremely careful not to start reading it until the upgrade has been confirmed, otherwise if the upgrade was rejected and we started forwarding the incoming request body, it would desynchronize the connection (potential security issue).
+				# We mitigate this issue by setting @persistent to false, which will prevent the connection from being reused, even if the upgrade fails (potential performance issue).
+				read_remainder_body
+			end
+			
 			HEAD = "HEAD"
 			CONNECT = "CONNECT"
 			
+			VALID_CONTENT_LENGTH = /\A\d+\z/
+			
 			def extract_content_length(headers)
 				if content_length = headers.delete(CONTENT_LENGTH)
-					if length = Integer(content_length, exception: false) and length >= 0
-						yield length
+					if content_length =~ VALID_CONTENT_LENGTH
+						yield Integer(content_length, 10)
 					else
-						raise BadRequest, "Invalid content length: #{content_length}"
+						raise BadRequest, "Invalid content length: #{content_length.inspect}"
 					end
 				end
 			end
@@ -428,6 +670,10 @@ module Protocol
 					return nil
 				end
 				
+				if status == 101
+					return read_upgrade_body
+				end
+				
 				if (status >= 100 and status < 200) or status == 204 or status == 304
 					return nil
 				end
@@ -454,6 +700,11 @@ module Protocol
 					return read_tunnel_body
 				end
 				
+				# A successful upgrade response implies that the connection will become a tunnel immediately after the empty line that concludes the header fields.
+				if headers[UPGRADE]
+					return read_upgrade_body
+				end
+				
 				# 6.  If this is a request message and none of the above are true, then
 				# the message body length is zero (no message body is present).
 				return read_body(headers)
